#
# Copyright (c), 2015-2020, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# Authors: Davide Brunato
#

###
# XML Schema namespaces

XSD_NAMESPACE = 'http://www.w3.org/2001/XMLSchema'
"URI of the XML Schema Definition namespace (xs|xsd)"

XSI_NAMESPACE = 'http://www.w3.org/2001/XMLSchema-instance'
"URI of the XML Schema Instance namespace (xsi)"

XML_NAMESPACE = 'http://www.w3.org/XML/1998/namespace'
"URI of the XML namespace (xml)"

###
# QE namespaces

PW_NAMESPACE = "http://www.quantum-espresso.org/ns/qes/qes-1.0"
"URI of the Quantum Espresso namespace for PW application"

PHONON_NAMESPACE = "http://www.quantum-espresso.org/ns/qes/qes_ph_1.0"
"URI of the Quantum Espresso namespace for Phonon application"

NEB_NAMESPACE = "http://www.quantum-espresso.org/ns/neb"
"URI of the Quantum Espresso namespace for NEB application"

TD_NAMESPACE = "http://www.quantum-espresso.org/ns/neb"
"URI of the Quantum Espresso namespace for TDDFPT application"

TD_SPECTRUM_NAMESPACE = "http://www.quantum-espresso.org/ns/qes/qes_spectrum-1.0"
"URI of the Quantum Espresso namespace for turbo-spectrum application"

# XSpectra namespace
XSPECTRA_NAMESPACE = "http://www.quantum-espresso.org/ns/qes/qes_xspectra-1.0"
"URI of the Quantum Espresso namespace for XSPECTRA application"
